# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

sys.path.append(os.path.abspath(".."))

project = "sleep-eeg-processing"
copyright = "2023, Gennadiy Belonosov"
author = "Gennadiy Belonosov"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
]

# autosummary_generate = False
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

autodoc_member_order = "bysource"
# autoclass_content = "class"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]



"""This module contains and describes pipe elements for sleep eeg analysis.
"""

from attrs import define, field
from pathlib import Path
import matplotlib.pyplot as plt
import numpy as np
import mne.io


from collections.abc import Iterable

from sleepeeg.base import BasePipe, BaseHypnoPipe, BaseEventPipe, BaseSpectrum


@define(kw_only=True)
class CleaningPipe(BasePipe):
    """The cleaning pipeline element.

    Contains resampling function, band and notch filters,
    browser for manual selection of bad channels
    and bad data spans.
    """

    def resample(
        self,
        save: bool = False,
        mne_resample_args: dict = None,
    ):
        """A wrapper for
        `mne.io.Raw.resample <https://mne.tools/stable/generated/mne.io.Raw.html#mne.io.Raw.resample>`_
        with an additional option to save the resampled data.

        Args:
            save: Whether to save a resampled data to a fif file. Defaults to False.
            mne_resample_args: Arguments passed to `mne.io.Raw.resample <https://mne.tools/stable/
                generated/mne.io.Raw.html#mne.io.Raw.resample>`_. Defaults to None.
        """
        mne_resample_args = mne_resample_args or dict()
        mne_resample_args.setdefault("sfreq", 250)
        self.mne_raw.resample(**mne_resample_args)
        if save:
            self.save_raw(
                "_".join(
                    filter(
                        None,
                        [
                            "resampled",
                            str(mne_resample_args["sfreq"]) + "hz",
                            "raw.fif",
                        ],
                    )
                )
            )

    def filter(
        self,
        mne_filter_args: dict = None,
    ):
        """A wrapper for
        `mne.io.Raw.filter <https://mne.tools/stable/generated/mne.io.Raw.html#mne.io.Raw.filter>`_.

        Args:
            mne_filter_args: Arguments passed to `mne.io.Raw.filter <https://mne.tools/stable/generated/mne.io.Raw.html#mne.io.Raw.filter>`_.
        """
        mne_filter_args = mne_filter_args or dict()
        mne_filter_args.setdefault("l_freq", 0.3)
        self.mne_raw.load_data()
        self.mne_raw.filter(**mne_filter_args)

    def notch(
        self,
        mne_notch_args: dict = None,
    ):
        """A wrapper for
        `mne.io.Raw.notch_filter <https://mne.tools/stable/generated/mne.io.Raw.html#mne.io.Raw.notch_filter>`_.

        Args:
            freqs: Frequencies to filter out from data,
                e.g. np.arange(50, 251, 50) for sampling freq 250 Hz.
                Defaults to None.
            picks: Channels to filter, if None - all channels will be filtered.
                Defaults to "eeg".
            n_jobs: The number of jobs to run in parallel or CUDA. Defaults to "cuda".
        """
        mne_notch_args = mne_notch_args or dict()
        mne_notch_args.setdefault("freqs", np.arange(50, int(self.sf / 2), 50))
        self.mne_raw.notch_filter(**mne_notch_args)

    def read_bad_channels(self, path=None):
        """Imports bad channels from file to mne raw object.

        Args:
            path: Path to the txt file with bad channel name per row. Defaults to None.
        """
        p = (
            Path(path)
            if path
            else self.output_dir / self.__class__.__name__ / "bad_channels.txt"
        )
        with open(p, "r") as f:
            self.mne_raw.info["bads"] = list(filter(None, f.read().split("\n")))

    def read_annotations(self, path=None):
        """Imports annotations from file to mne raw object

        Args:
            path: Path to txt file with mne-style annotations. Defaults to None.
        """
        from mne import read_annotations

        p = (
            Path(path)
            if path
            else self.output_dir / self.__class__.__name__ / "annotations.txt"
        )
        self.mne_raw.set_annotations(read_annotations(p))


@define(kw_only=True)
class ICAPipe(BasePipe):
    """The ICA pipeline element.

    Contains ica fitting, plotting multiple ica plots,
    selecting ica exclusion components and
    its application to the raw data.
    More at `mne.preprocessing.ICA <https://mne.tools/stable/
    generated/mne.preprocessing.ICA.html#mne-preprocessing-ica>`_.
    """

    n_components: int | float | None = field(default=30)
    """Number of principal components (from the pre-whitening PCA step) 
    that are passed to the ICA algorithm during fitting.
    """

    method: str = field(default="fastica")
    """The ICA method to use in the fit method. 

    Can be 'fastica', 'infomax' or 'picard'
    Use the fit_params argument to set additional parameters.
    """
    fit_params: dict = field(default=None)
    """Additional parameters passed to the ICA estimator as specified by method. 
    Allowed entries are determined by the various algorithm implementations: 
    see `FastICA <https://scikit-learn.org/stable/modules/generated/sklearn.
    decomposition.FastICA.html#sklearn.decomposition.FastICA>`_, 
    `picard <https://pierreablin.github.io/picard/generated/picard.picard.html#picard.picard>`_ and
    `infomax <https://mne.tools/stable/generated/mne.preprocessing.infomax.html#mne.preprocessing.infomax>`_.
    """

    path_to_ica: Path = field(converter=Path, default="/")

    mne_ica: mne.preprocessing.ICA = field()
    """Instance of 
    `mne.preprocessing.ICA <https://mne.tools/stable/
    generated/mne.preprocessing.ICA.html#mne-preprocessing-ica>`_.
    """

    @mne_ica.default
    def _set_mne_ica(self):
        if self.path_to_ica == Path("/"):
            return mne.preprocessing.ICA(
                n_components=self.n_components,
                method=self.method,
                fit_params=self.fit_params,
            )
        return mne.preprocessing.read_ica(self.path_to_ica)

    def __attrs_post_init__(self):
        self.mne_raw.load_data()

    def fit(self, filter_args=None, ica_fit_args=None):
        """Highpass-filters (1Hz) a copy of the mne_raw object
        and then runs `mne.preprocessing.ICA.fit <https://mne.tools/stable/
        generated/mne.preprocessing.ICA.html#mne.preprocessing.ICA.fit>`_.
        """
        filter_args = filter_args or dict()
        ica_fit_args = ica_fit_args or dict()
        filter_args.setdefault("l_freq", 1.0)
        if self.mne_raw.info["highpass"] < 1.0:
            filtered_raw = self.mne_raw.copy()
            filtered_raw.filter(**filter_args)
        else:
            filtered_raw = self.mne_raw
        self.mne_ica.fit(filtered_raw, **ica_fit_args)

    def plot_sources(self, **kwargs):
        """A wrapper for `mne.preprocessing.ICA.plot_sources <https://mne.tools/stable/
        generated/mne.preprocessing.ICA.html#mne.preprocessing.ICA.plot_sources>`_.
        """
        self.mne_ica.plot_sources(self.mne_raw, block=True, **kwargs)

    def plot_components(self, **kwargs):
        """A wrapper for `mne.preprocessing.ICA.plot_components <https://mne.tools/stable/
        generated/mne.preprocessing.ICA.html#mne.preprocessing.ICA.plot_components>`_.
        """
        self.mne_ica.plot_components(inst=self.mne_raw, **kwargs)

    def plot_overlay(self, exclude=None, picks=None, start=10, stop=20, **kwargs):
        """A wrapper for `mne.preprocessing.ICA.plot_overlay <https://mne.tools/stable/
        generated/mne.preprocessing.ICA.html#mne.preprocessing.ICA.plot_overlay>`_.
        """
        self.mne_ica.plot_overlay(
            self.mne_raw, exclude=exclude, picks=picks, start=start, stop=stop, **kwargs
        )

    def plot_properties(self, picks=None, **kwargs):
        """A wrapper for `mne.preprocessing.ICA.plot_properties <https://mne.tools/stable/
        generated/mne.preprocessing.ICA.html#mne.preprocessing.ICA.plot_properties>`_.
        """
        self.mne_ica.plot_properties(self.mne_raw, picks=picks, **kwargs)

    def apply(self, exclude=None, **kwargs):
        """Remove selected components from the signal.

        A wrapper for `mne.preprocessing.ICA.apply <https://mne.tools/stable/
        generated/mne.preprocessing.ICA.html#mne.preprocessing.ICA.apply>`_.
        """
        self.mne_ica.apply(self.mne_raw, exclude=exclude, **kwargs)

    def save_ica(self, fname="data-ica.fif", overwrite=False):
        """A wrapper for `mne.preprocessing.ICA.save <https://mne.tools/stable/
        generated/mne.preprocessing.ICA.html#mne.preprocessing.ICA.save>`_.

        Args:
            fname: filename for the ica file being saved. Defaults to "data-ica.fif".
            overwrite: Whether to overwrite the file. Defaults to False.
        """
        fif_folder = self.output_dir / self.__class__.__name__ / "saved_ica"
        fif_folder.mkdir(exist_ok=True)
        self.mne_ica.save(fif_folder / fname, overwrite=overwrite)


@define(kw_only=True)
class SpectralPipe(BaseHypnoPipe, BaseSpectrum):
    """The spectral analyses pipeline element.

    Contains methods for computing and plotting PSD,
    spectrogram and topomaps, per sleep stage.
    """

    def plot_hypnospectrogram(
        self,
        picks: str | Iterable[str] = ("E101",),
        sec_per_seg: float = 4.096,
        freq_range: tuple = (0, 40),
        cmap: str = "inferno",
        overlap: bool = False,
        save: bool = False,
    ):
        """Plots hypnogram and spectrogram.

        Args:
            picks: Channels to calculate spectrogram on, more info at
                `mne.io.Raw.get_data <https://mne.tools/stable/generated/
                mne.io.Raw.html#mne.io.Raw.get_data>`_.
                Defaults to ("E101",).
            sec_per_seg: Segment length in seconds. Defaults to 4.096.
            freq_range: Range of x axis on spectrogram plot. Defaults to (0, 40).
            cmap: Matplotlib `colormap <https://matplotlib.org/stable/tutorials/colors/colormaps.html>`_. Defaults to "inferno".
            overlap: Whether to plot hypnogram over the spectrogram or on top of it. Defaults to False.
            save: Whether to save the figure. Defaults to False.
        """
        # Import data from the raw mne file.
        data = self.mne_raw.get_data(picks, units="uV", reject_by_annotation="NaN")[0]
        # Create a plot figure
        fig = self.__plot_hypnospectrogram(
            data,
            self.sf,
            self.hypno_up,
            win_sec=sec_per_seg,
            fmin=freq_range[0],
            fmax=freq_range[1],
            trimperc=0,
            cmap=cmap,
            overlap=overlap,
        )
        # Save the figure if 'save' set to True
        if save:
            fig.savefig(
                self.output_dir / self.__class__.__name__ / f"spectrogram.png",
                bbox_inches="tight",
            )

    def _compute_psd_per_stage(
        self, picks, sleep_stages, avg_ref, dB, method_args=None
    ):
        from collections import defaultdict
        from scipy import signal, ndimage

        assert (
            self.hypno.any()
        ), f"Hypnogram hasn't been provided, can't compute PSD per stage"

        method_args = method_args or dict()
        method_args["axis"] = 1
        if "nperseg" not in method_args:
            method_args["nperseg"] = self.sf * 4.096
        # Import data from the raw mne file.
        self.mne_raw.load_data()
        if avg_ref:
            data = (
                self.mne_raw.copy()
                .set_eeg_reference()
                .get_data(picks=picks, units="uV", reject_by_annotation="NaN")
            )
        else:
            data = self.mne_raw.get_data(
                picks=picks, units="uV", reject_by_annotation="NaN"
            )
        data = np.ma.array(data, mask=np.isnan(data))
        n_samples_total = np.ma.compress_cols(data).shape[1]
        psds = defaultdict(list)
        psd_per_stage = {}
        for stage, index in sleep_stages.items():
            weights = []
            n_samples = 0
            try:
                regions = ndimage.find_objects(
                    ndimage.label(
                        np.logical_or.reduce([self.hypno_up == i for i in index])
                    )[0]
                )
            except TypeError:
                regions = ndimage.find_objects(ndimage.label(self.hypno_up == index)[0])
            for region in regions:
                compressed = np.ma.compress_cols(data[:, region[0]])
                if compressed.size > method_args["nperseg"]:
                    weights.append(compressed.shape[1])
                    freqs, psd = signal.welch(
                        compressed,
                        fs=self.sf,
                        **method_args,
                    )
                    psds[stage].append(psd)
                    n_samples += compressed.shape[1]
            avg = np.average(np.array(psds[stage]), weights=weights, axis=0)
            psd_per_stage[stage] = [
                freqs,
                10 * np.log10(avg) if dB else avg,
                round(n_samples / n_samples_total * 100, 2),
            ]
        return psd_per_stage

    def __plot_hypnospectrogram(
        self, data, sf, hypno, win_sec, fmin, fmax, trimperc, cmap, overlap
    ):
        """
        ?
        """
        # Increase font size while preserving original
        old_fontsize = plt.rcParams["font.size"]
        plt.rcParams.update({"font.size": 18})

        if overlap or not hypno.any():
            fig, ax = plt.subplots(nrows=1, figsize=(12, 4))
            im = self.__plot_spectrogram(
                data, sf, win_sec, fmin, fmax, trimperc, cmap, ax
            )
            if hypno.any():
                ax_hypno = ax.twinx()
                self.__plot_hypnogram(sf, hypno, ax_hypno)
            # Add colorbar
            cbar = fig.colorbar(
                im, ax=ax, shrink=0.95, fraction=0.1, aspect=25, pad=0.1
            )
            cbar.ax.set_ylabel("Log Power (dB / Hz)", rotation=90, labelpad=20)
            # Revert font-size
            plt.rcParams.update({"font.size": old_fontsize})
            return fig

        fig, (ax0, ax1) = plt.subplots(
            nrows=2, figsize=(12, 6), gridspec_kw={"height_ratios": [1, 2]}
        )
        plt.subplots_adjust(hspace=0.1)

        # Hypnogram (top axis)
        self.__plot_hypnogram(sf, hypno, ax0)
        # Spectrogram (bottom axis)
        self.__plot_spectrogram(data, sf, win_sec, fmin, fmax, trimperc, cmap, ax1)
        # Revert font-size
        plt.rcParams.update({"font.size": old_fontsize})
        return fig

    @staticmethod
    def __plot_hypnogram(sf, hypno, ax0):
        from pandas import Series

        hypno = np.asarray(hypno).astype(int)
        t_hyp = np.arange(hypno.size) / (sf * 3600)
        # Make sure that REM is displayed after Wake
        hypno = Series(hypno).map({-2: -2, -1: -1, 0: 0, 1: 2, 2: 3, 3: 4, 4: 1}).values
        # Hypnogram (top axis)
        ax0.step(t_hyp, -1 * hypno, color="k")
        if -2 in hypno and -1 in hypno:
            # Both Unscored and Artefacts are present
            ax0.set_yticks([2, 1, 0, -1, -2, -3, -4])
            ax0.set_yticklabels(["Uns", "Art", "W", "R", "N1", "N2", "N3"])
            ax0.set_ylim(-4.5, 2.5)
        elif -2 in hypno and -1 not in hypno:
            # Only Unscored are present
            ax0.set_yticks([2, 0, -1, -2, -3, -4])
            ax0.set_yticklabels(["Uns", "W", "R", "N1", "N2", "N3"])
            ax0.set_ylim(-4.5, 2.5)

        elif -2 not in hypno and -1 in hypno:
            # Only Artefacts are present
            ax0.set_yticks([1, 0, -1, -2, -3, -4])
            ax0.set_yticklabels(["Art", "W", "R", "N1", "N2", "N3"])
            ax0.set_ylim(-4.5, 1.5)
        else:
            # No artefacts or Unscored
            ax0.set_yticks([0, -1, -2, -3, -4])
            ax0.set_yticklabels(["W", "R", "N1", "N2", "N3"])
            ax0.set_ylim(-4.5, 0.5)
        ax0.set_xlim(0, t_hyp.max())
        ax0.set_ylabel("Stage")
        ax0.xaxis.set_visible(False)
        ax0.spines["right"].set_visible(False)
        ax0.spines["top"].set_visible(False)
        return ax0

    @staticmethod
    def __plot_spectrogram(data, sf, win_sec, fmin, fmax, trimperc, cmap, ax):
        from matplotlib.colors import Normalize
        from lspopt import spectrogram_lspopt
        import numpy as np

        # Calculate multi-taper spectrogram
        nperseg = int(win_sec * sf)
        f, t, Sxx = spectrogram_lspopt(data, sf, nperseg=nperseg, noverlap=0)
        Sxx = 10 * np.log10(
            Sxx, out=np.full(Sxx.shape, np.nan), where=(Sxx != 0)
        )  # Convert uV^2 / Hz --> dB / Hz

        # Select only relevant frequencies (up to 30 Hz)
        good_freqs = np.logical_and(f >= fmin, f <= fmax)
        Sxx = Sxx[good_freqs, :]
        f = f[good_freqs]
        t /= 3600  # Convert t to hours

        # Normalization
        vmin, vmax = np.nanpercentile(Sxx, [0 + trimperc, 100 - trimperc])
        norm = Normalize(vmin=vmin, vmax=vmax)
        im = ax.pcolormesh(
            t, f, Sxx, norm=norm, cmap=cmap, antialiased=True, shading="auto"
        )
        ax.set_xlim(0, t.max())
        ax.set_ylabel("Frequency [Hz]")
        ax.set_xlabel("Time [hrs]")
        return im


@define(kw_only=True)
class SpindlesPipe(BaseEventPipe):
    """Spindles detection."""

    def detect(
        self,
        picks: str | Iterable[str] = ("eeg"),
        include: Iterable[int] = (1, 2, 3),
        freq_sp: Iterable[float] = (12, 15),
        freq_broad: Iterable[float] = (1, 30),
        duration: Iterable[float] = (0.5, 2),
        min_distance: int = 500,
        thresh: dict = {"corr": 0.65, "rel_pow": 0.2, "rms": 1.5},
        multi_only: bool = False,
        remove_outliers: bool = False,
        verbose: bool = False,
        save: bool = False,
    ):
        """Wrapper around YASA's `spindles_detect <https://raphaelvallat.com/
        yasa/build/html/generated/yasa.spindles_detect.html>`_.
        """
        from yasa import spindles_detect

        self.results = spindles_detect(
            data=self.mne_raw.copy().load_data().set_eeg_reference().pick(picks),
            hypno=self.hypno_up,
            verbose=verbose,
            include=include,
            freq_sp=freq_sp,
            freq_broad=freq_broad,
            duration=duration,
            min_distance=min_distance,
            thresh=thresh,
            multi_only=multi_only,
            remove_outliers=remove_outliers,
        )
        if save:
            self._save_to_csv()


@define(kw_only=True)
class SlowWavesPipe(BaseEventPipe):
    """Slow waves detection."""

    def detect(
        self,
        picks: str | Iterable[str] = ("eeg"),
        include: Iterable[int] = (1, 2, 3),
        freq_sw: Iterable[float] = (0.3, 1.5),
        dur_neg: Iterable[float] = (0.3, 1.5),
        dur_pos: Iterable[float] = (0.1, 1),
        amp_neg: Iterable[float] = (40, 200),
        amp_pos: Iterable[float] = (10, 150),
        amp_ptp: Iterable[float] = (75, 350),
        coupling: bool = False,
        coupling_params: dict = {"freq_sp": (12, 16), "p": 0.05, "time": 1},
        remove_outliers: bool = False,
        save: bool = False,
    ):
        """Wrapper around YASA's `sw_detect <https://raphaelvallat.com/yasa/
        build/html/generated/yasa.sw_detect.html>`_.
        """
        from yasa import sw_detect

        self.results = sw_detect(
            data=self.mne_raw.copy().load_data().set_eeg_reference().pick(picks),
            hypno=self.hypno_up,
            verbose=False,
            include=include,
            freq_sw=freq_sw,
            dur_neg=dur_neg,
            dur_pos=dur_pos,
            amp_neg=amp_neg,
            amp_pos=amp_pos,
            amp_ptp=amp_ptp,
            coupling=coupling,
            coupling_params=coupling_params,
            remove_outliers=remove_outliers,
        )
        if save:
            self._save_to_csv()


@define(kw_only=True)
class REMsPipe(BaseEventPipe):
    """Rapid eye movements detection."""

    def detect(
        self,
        loc_chname: str = "E46",
        roc_chname: str = "E238",
        include: int | Iterable[int] = 4,
        freq_rem: Iterable[float] = (0.5, 5),
        duration: Iterable[float] = (0.3, 1.2),
        amplitude: Iterable[float] = (50, 325),
        remove_outliers: bool = False,
        save: bool = False,
    ):
        """Wrapper around YASA's `rem_detect <https://raphaelvallat.com/yasa/
        build/html/generated/yasa.rem_detect.html>`_.
        """
        from yasa import rem_detect

        referenced = self.mne_raw.copy().load_data().set_eeg_reference()
        loc = referenced.get_data([loc_chname], units="uV", reject_by_annotation="NaN")
        roc = referenced.get_data([roc_chname], units="uV", reject_by_annotation="NaN")
        self.results = rem_detect(
            loc=loc,
            roc=roc,
            sf=self.sf,
            hypno=self.hypno_up,
            verbose=False,
            include=include,
            freq_rem=freq_rem,
            duration=duration,
            amplitude=amplitude,
            remove_outliers=remove_outliers,
        )
        if save:
            self._save_to_csv()

    def plot_average(self, save=False, yasa_args=None):
        yasa_args = yasa_args or dict()
        self.results.plot_average(**yasa_args)
        if save:
            self._save_avg_fig()

    def plot_topomap(self):
        raise AttributeError("'REMsPipe' object has no attribute 'plot_topomap'")

    def plot_topomap_collage(self):
        raise AttributeError("'REMsPipe' object has no attribute 'plot_topomap'")


@define(kw_only=True)
class CombinedPipe(BaseSpectrum):
    """The pipeline element combining results from multiple subjects.

    Contains methods for computing and plotting combined PSD,
    spectrogram and topomaps, per sleep stage.
    """

    pipes: Iterable[BaseHypnoPipe] = field()
    """Stores pipeline elements for multiple subjects.
    """

    mne_raw: mne.io.Raw = field(init=False)

    @mne_raw.default
    def _get_mne_raw_from_pipe(self):
        return self.pipes[0].mne_raw

    def _compute_psd_per_stage(self, picks, sleep_stages, sec_per_seg, avg_ref, dB):
        psd_per_stage = {}
        psds = []
        for pipe in self.pipes:
            psds.append(
                pipe._compute_psd_per_stage(
                    picks=picks,
                    sleep_stages=sleep_stages,
                    sec_per_seg=sec_per_seg,
                    avg_ref=avg_ref,
                    dB=dB,
                )
            )

        for stage in sleep_stages:
            psd_per_stage[stage] = [
                psds[0][stage][0],
                np.sum([psd[stage][1] for psd in psds], axis=0) / len(self.pipes),
                round(sum([psd[stage][2] for psd in psds]) / len(self.pipes), 2),
            ]
        return psd_per_stage



import os
import errno

from abc import ABC, abstractmethod
from collections.abc import Iterable
from pathlib import Path

from typing import TypeVar, Type
from attrs import define, field

import matplotlib.pyplot as plt
import numpy as np
import mne
from tqdm import tqdm

# For type annotation of pipe elements.
BasePipeType = TypeVar("BasePipeType", bound="BasePipe")


@define(kw_only=True, slots=False)
class BasePipe(ABC):
    """A template class for the pipeline segments."""

    prec_pipe: Type[BasePipeType] = field(default=None, metadata={"my_metadata": 1})
    """Preceding pipe that hands over mne_raw attr."""

    path_to_eeg: Path = field(converter=Path)
    """Can be any file type supported by 
    `mne.io.read_raw() <https://mne.tools/stable/generated/
    mne.io.Raw.html#mne.io.Raw.save>`_.
    """

    @path_to_eeg.default
    def _set_path_to_eeg(self):
        if self.prec_pipe:
            return "/"
        raise TypeError('Provide either "pipe" or "path_to_eeg" arguments')

    @path_to_eeg.validator
    def _validate_path_to_eeg(self, attr, value):
        if not value.exists():
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), value)

    output_dir: Path = field(converter=Path)
    """Path to the directory where the output will be saved."""

    @output_dir.default
    def _set_output_dir(self):
        return (
            self.prec_pipe.output_dir if self.prec_pipe else self.path_to_eeg.parents[0]
        )

    @output_dir.validator
    def _validate_output_dir(self, attr, value):
        self.output_dir.mkdir(exist_ok=True)
        (self.output_dir / self.__class__.__name__).mkdir(exist_ok=True)

    mne_raw: mne.io.Raw = field(init=False)
    """An instanse of  
    `mne.io.Raw <https://mne.tools/stable/generated/
    mne.io.Raw.html#mne-io-raw>`_.
    """

    @mne_raw.default
    def _read_mne_raw(self):
        from mne.io import read_raw

        try:
            return (
                self.prec_pipe.mne_raw if self.prec_pipe else read_raw(self.path_to_eeg)
            )
        except Exception as err:
            print(f"Unexpected {err=}, {type(err)=}")
            raise

    @property
    def sf(self):
        """A wrapper for
        `mne.Info["sfreq"] <https://mne.tools/stable/generated/
        mne.Info.html#mne-info>`_.

        Returns:
            float: sampling frequency
        """
        return self.mne_raw.info["sfreq"]

    def plot(
        self,
        save_annotations: bool = False,
        save_bad_channels: bool = False,
        overwrite: bool = False,
        mne_plot_args: dict = None,
    ):
        """A wrapper for `mne.io.Raw.plot() <https://mne.tools/stable/
        generated/mne.io.Raw.html#mne.io.Raw.plot>`_.

        Args:
            butterfly: Whether to start in butterfly mode. Defaults to False.
            save_annotations: Whether to save annotations as txt. Defaults to False.
            save_bad_channels: Whether to save bad channels as txt. Defaults to False.
            scalings: Scale for amplitude per channel type. Defaults to 'auto'.
            use_opengl: Whether to use OpenGL acceleration. Defaults to None.
            overwrite: Whether to overwrite annotations and bad_channels files if exist.
                Defaults to False.
            mne_plot_args: Arguments passed to `raw.plot() <https://mne.tools/stable/
                generated/mne.io.Raw.html#mne.io.Raw.plot>`_. Defaults to None.
        """
        mne_plot_args = mne_plot_args or dict()
        mne_plot_args.setdefault("theme", "dark")
        mne_plot_args.setdefault("block", True)
        mne_plot_args.setdefault("bad_color", "r")
        mne_plot_args.setdefault("scalings", "auto")
        self.mne_raw.plot(**mne_plot_args)

        if save_annotations:
            self.mne_raw.annotations.save(
                self.output_dir / self.__class__.__name__ / "annotations.txt",
                overwrite=overwrite,
            )
        if save_bad_channels:
            if overwrite:
                with open(
                    self.output_dir / self.__class__.__name__ / "bad_channels.txt", "w"
                ) as f:
                    for bad in self.mne_raw.info["bads"]:
                        f.write(f"{bad}\n")
            else:
                with open(
                    self.output_dir / self.__class__.__name__ / "bad_channels.txt", "a+"
                ) as f:
                    f.seek(0)
                    bads = f.read().split()
                    for bad in self.mne_raw.info["bads"]:
                        if bad not in bads:
                            f.write(f"{bad}\n")

    def _plot_sensors(
        self,
        ch_colors=None,
        kind="topomap",
        ch_type=None,
        title=None,
        show_names=False,
        ch_groups=None,
        to_sphere=True,
        axes=None,
        block=False,
        show=False,
        sphere=None,
        pointsize=7,
        linewidth=1,
    ):
        from mne.viz.evoked import _rgb
        from mne.viz.utils import _plot_sensors
        from mne.defaults import _handle_default
        from mne.io.constants import FIFF
        from mne.io.pick import (
            channel_type,
            channel_indices_by_type,
            pick_channels,
            _DATA_CH_TYPES_SPLIT,
            _contains_ch_type,
        )
        from mne.utils import _check_ch_locs, _check_option, warn
        from mne.transforms import apply_trans

        info = self.mne_raw.info
        _check_option("kind", kind, ["topomap", "3d", "select"])
        if not isinstance(info, mne.io.Info):
            raise TypeError(f"info must be an instance of Info not {type(info)}")
        ch_indices = channel_indices_by_type(info)
        allowed_types = _DATA_CH_TYPES_SPLIT
        if ch_type is None:
            for this_type in allowed_types:
                if _contains_ch_type(info, this_type):
                    ch_type = this_type
                    break
            picks = ch_indices[ch_type]
        elif ch_type == "all":
            picks = list()
            for this_type in allowed_types:
                picks += ch_indices[this_type]
        elif ch_type in allowed_types:
            picks = ch_indices[ch_type]
        else:
            raise ValueError(f"ch_type must be one of {allowed_types} not {ch_type}!")

        if len(picks) == 0:
            raise ValueError(f"Could not find any channels of type {ch_type}.")

        if not _check_ch_locs(info=info, picks=picks):
            raise RuntimeError("No valid channel positions found")

        dev_head_t = info["dev_head_t"]
        chs = [info["chs"][pick] for pick in picks]
        pos = np.empty((len(chs), 3))
        for ci, ch in enumerate(chs):
            pos[ci] = ch["loc"][:3]
            if ch["coord_frame"] == FIFF.FIFFV_COORD_DEVICE:
                if dev_head_t is None:
                    warn(
                        "dev_head_t is None, transforming MEG sensors to head "
                        "coordinate frame using identity transform"
                    )
                    dev_head_t = np.eye(4)
                pos[ci] = apply_trans(dev_head_t, pos[ci])
        del dev_head_t

        ch_names = np.array([ch["ch_name"] for ch in chs])
        bads = [idx for idx, name in enumerate(ch_names) if name in info["bads"]]
        ch_colors = {} if not ch_colors else ch_colors
        if ch_groups is None:
            def_colors = _handle_default("color")
            colors = [
                ch_colors[self.mne_raw.info.ch_names[pick]]
                if self.mne_raw.info.ch_names[pick] in ch_colors
                else def_colors[channel_type(info, pick)]
                for i, pick in enumerate(picks)
            ]
        else:
            if ch_groups in ["position", "selection"]:
                # Avoid circular import
                from mne.channels import (
                    read_vectorview_selection,
                    _SELECTIONS,
                    _EEG_SELECTIONS,
                    _divide_to_regions,
                )

                if ch_groups == "position":
                    ch_groups = _divide_to_regions(info, add_stim=False)
                    ch_groups = list(ch_groups.values())
                else:
                    ch_groups, color_vals = list(), list()
                    for selection in _SELECTIONS + _EEG_SELECTIONS:
                        channels = pick_channels(
                            info["ch_names"],
                            read_vectorview_selection(selection, info=info),
                        )
                        ch_groups.append(channels)
                color_vals = np.ones((len(ch_groups), 4))
                for idx, ch_group in enumerate(ch_groups):
                    color_picks = [
                        np.where(picks == ch)[0][0] for ch in ch_group if ch in picks
                    ]
                    if len(color_picks) == 0:
                        continue
                    x, y, z = pos[color_picks].T
                    color = np.mean(_rgb(x, y, z), axis=0)
                    color_vals[idx, :3] = color  # mean of spatial color
            else:
                import matplotlib.pyplot as plt

                colors = np.linspace(0, 1, len(ch_groups))
                color_vals = [plt.cm.jet(colors[i]) for i in range(len(ch_groups))]
            if not isinstance(ch_groups, (np.ndarray, list)):
                raise ValueError(
                    "ch_groups must be None, 'position', "
                    "'selection', or an array. Got %s." % ch_groups
                )
            colors = np.zeros((len(picks), 4))
            for pick_idx, pick in enumerate(picks):
                for ind, value in enumerate(ch_groups):
                    if pick in value:
                        colors[pick_idx] = color_vals[ind]
                        break
        title = "Sensor positions (%s)" % ch_type if title is None else title
        fig = _plot_sensors(
            pos,
            info,
            picks,
            colors,
            bads,
            ch_names,
            title,
            show_names,
            axes,
            show,
            kind,
            block,
            to_sphere,
            sphere,
            pointsize=pointsize,
            linewidth=linewidth,
        )

        if kind == "select":
            return fig, fig.lasso.selection
        return fig

    def save_raw(self, fname: str):
        """A wrapper for `mne.io.Raw.save <https://mne.tools/stable/
        generated/mne.io.Raw.html#mne.io.Raw.save>`_.

        Args:
            fname: filename for the fif file being saved.
        """
        fif_folder = self.output_dir / self.__class__.__name__
        self.mne_raw.save(fif_folder / fname)


@define(kw_only=True, slots=False)
class BaseHypnoPipe(BasePipe, ABC):
    """A template class for the sleep stage analysis pipeline segments."""

    path_to_hypno: Path = field(converter=Path)
    """Path to hypnogram. Must be text file with every 
    row being int representing sleep stage for the epoch.
    """

    @path_to_hypno.default
    def _set_path_to_hypno(self):
        return "/"

    @path_to_hypno.validator
    def _validate_path_to_hypno(self, attr, value):
        if not value.exists():
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), value)

    hypno_freq: float = field(converter=float)
    """Sampling rate of the hypnogram in Hz.

    E.g., 1/30 means 1 sample per 30 secs epoch,
    250 means 1 sample per 1/250 sec epoch.
    """

    @hypno_freq.default
    def _get_hypno_freq(self):
        if self.prec_pipe and isinstance(self.prec_pipe, BaseHypnoPipe):
            return self.prec_pipe.hypno_freq
        return 1

    hypno: np.ndarray = field()
    """ Hypnogram with sampling frequency hypno_freq
    with int representing sleep stage.
    """

    @hypno.default
    def _import_hypno(self):
        if self.prec_pipe and isinstance(self.prec_pipe, BaseHypnoPipe):
            return self.prec_pipe.hypno
        if self.path_to_hypno == Path("/"):
            return np.empty(0)
        try:
            return np.loadtxt(self.path_to_hypno)
        except Exception as err:
            print(f"Unexpected {err=}, {type(err)=}")
            raise

    hypno_up: np.array = field()
    """ Hypnogram upsampled to the sampling frequency of the raw data.
    """

    @hypno_up.default
    def _set_hypno_up(self):
        return self.hypno

    def __attrs_post_init__(self):
        if self.hypno.any():
            self.__upsample_hypno()

    def __upsample_hypno(self):
        from yasa import hypno_upsample_to_data

        self.hypno_up = hypno_upsample_to_data(
            self.hypno, self.hypno_freq, self.mne_raw, verbose=False
        )

    def predict_hypno(
        self,
        eeg_name: str = "E183",
        eog_name: str = "E252",
        emg_name: str = "E247",
        ref_name: str = "E26",
        save=True,
    ):
        """Runs YASA's automatic sleep staging

        Args:
            eeg_name: Preferentially a central electrode. Defaults to "E183".
            eog_name: Preferentially, the left LOC channel. Defaults to "E252".
            emg_name: Preferentially a chin electrode. Defaults to "E247".
            ref_name: Reference channel, preferentially a mastoid. Defaults to "E26".
            save: Whether to save the hypnogram. Defaults to True.
        """
        from yasa import SleepStaging, hypno_str_to_int

        sls = SleepStaging(
            self.mne_raw.copy().load_data().set_eeg_reference(ref_channels=[ref_name]),
            eeg_name=eeg_name,
            eog_name=eog_name,
            emg_name=emg_name,
        )
        hypno = sls.predict()
        self.hypno = hypno_str_to_int(hypno)
        self.hypno_freq = 1 / 30
        self.__upsample_hypno()
        sls.plot_predict_proba()
        if save:
            np.savetxt(
                self.output_dir / self.__class__.__name__ / "predicted_hypno.txt",
                self.hypno,
                fmt="%d",
            )
            plt.savefig(
                self.output_dir
                / self.__class__.__name__
                / "predicted_hypno_probabilities.png"
            )

    def sleep_stats(self, save: bool = False):
        """A wrapper for
        `yasa.sleep_statistics <https://raphaelvallat.com/yasa/build/html/generated/yasa.sleep_statistics.html#yasa-sleep-statistics>`_.

        Args:
            save: Whether to save the stats to csv. Defaults to False.
        """

        from yasa import sleep_statistics
        from csv import DictWriter

        assert self.hypno.any(), "There is no hypnogram to get stats from."
        stats = sleep_statistics(self.hypno, self.hypno_freq)
        if save:
            with open(
                self.output_dir / self.__class__.__name__ / "sleep_stats.csv",
                "w",
                newline="",
            ) as csv_file:
                w = DictWriter(csv_file, stats.keys())
                w.writeheader()
                w.writerow(stats)
            return
        return stats


@define(kw_only=True, slots=False)
class BaseTopomap(ABC):
    def _plot_topomap(self, data, axis, info=None, topomap_args=None, cbar_args=None):
        from mne.viz import plot_topomap

        topomap_args = topomap_args or dict()
        topomap_args.setdefault("size", 5)
        topomap_args.setdefault("show", False)
        im, cn = plot_topomap(
            data,
            info if info else self.mne_raw.info,
            axes=axis,
            **topomap_args,
        )

        cbar_args = cbar_args or dict()
        cbar_args.setdefault("shrink", 0.6)
        cbar_args.setdefault("orientation", "vertical")
        plt.colorbar(
            im,
            ax=axis,
            **cbar_args,
        )


@define(kw_only=True, slots=False)
class BaseEventPipe(BaseHypnoPipe, BaseTopomap, ABC):
    """A template class for event detection."""

    results = field(init=False)
    """Event detection results as returned by YASA's event detection methods. 
    Depending on the child class can be instance of either 
    SpindlesResults, SWResults or REMResults classes. 
    """

    tfrs: dict = field(init=False)
    """Instances of mne.time_frequency.AverageTFR per sleep stage.
    """

    @abstractmethod
    def detect():
        pass

    def _save_to_csv(self):
        self.results.summary().to_csv(
            self.output_dir
            / self.__class__.__name__
            / f"{self.__class__.__name__[:-4].lower()}.csv",
            index=False,
        )

    def _save_avg_fig(self):
        plt.savefig(
            self.output_dir
            / self.__class__.__name__
            / f"{self.__class__.__name__[:-4].lower()}_avg.png"
        )

    def plot_average(self, save: bool = False, yasa_args=None):
        """Average of YASA's detected event.

        Args:
            save: Whether to save the figure to file. Defaults to False.
            yasa_args: Arguments passed to the YASA's plot_average(). Defaults to None.
        """
        yasa_args = yasa_args or dict()
        self.results.plot_average(**yasa_args)
        if save:
            self._save_avg_fig()

    def plot_topomap(
        self,
        prop: str,
        stage: str = "N2",
        aggfunc: str = "mean",
        sleep_stages: dict = {"Wake": 0, "N1": 1, "N2": 2, "N3": 3, "REM": 4},
        axis: plt.axis = None,
        save: bool = False,
        topomap_args: dict = None,
        cbar_args: dict = None,
        subplots_args: dict = None,
    ):
        """Plots topomap for a sleep stage and some property of detected events.

        Args:
            prop: Any event property returned by self.results.summary().
            stage: One of the sleep_stages keys. Defaults to "N2".
            aggfunc: Averaging function, "mean" or "median". Defaults to "mean".
            sleep_stages: Mapping between sleep stages names and their integer representations.
                Defaults to {"Wake": 0, "N1": 1, "N2": 2, "N3": 3, "REM": 4}.
            axis: Instance of `matplotlib.pyplot.axis <https://matplotlib.org/
                stable/api/_as_gen/matplotlib.pyplot.axis.html#matplotlib-pyplot-axis>`_.
                Defaults to None.
            cmap: Matplotlib `colormap <https://matplotlib.org/stable/tutorials/colors/colormaps.html>`_.
                Defaults to "plasma".
            save: Whether to save the figure. Defaults to False.
        """
        from natsort import natsort_keygen
        from more_itertools import collapse
        from seaborn import color_palette

        topomap_args = topomap_args or dict()
        cbar_args = cbar_args or dict()
        subplots_args = subplots_args or dict()
        topomap_args.setdefault("cmap", color_palette("rocket_r", as_cmap=True))
        cbar_args.setdefault("label", prop)

        grouped_summary = (
            self.results.summary(grp_chan=True, grp_stage=True, aggfunc=aggfunc)
            .sort_values("Channel", key=natsort_keygen())
            .reset_index()
        )
        assert np.isin(
            sleep_stages[stage], grouped_summary["Stage"].unique()
        ).all(), "No such stage in the detected events, was it included in the detect method?"

        is_new_axis = False
        if not axis:
            fig, axis = plt.subplots(**subplots_args)
            is_new_axis = True

        per_stage = grouped_summary.loc[
            grouped_summary["Stage"].isin(collapse([sleep_stages[stage]]))
        ].groupby("Channel")
        per_stage = per_stage.mean() if aggfunc == "mean" else per_stage.median()
        per_stage = per_stage.sort_values("Channel", key=natsort_keygen()).reset_index()

        info = self.mne_raw.copy().pick(list(per_stage["Channel"].unique())).info

        topomap_args.setdefault("vlim", (per_stage[prop].min(), per_stage[prop].max()))
        self._plot_topomap(
            data=per_stage[prop],
            axis=axis,
            info=info,
            topomap_args=topomap_args,
            cbar_args=cbar_args,
        )
        if is_new_axis:
            fig.suptitle(f"{stage} {self.__class__.__name__[:-4]} ({prop})")
        if save and is_new_axis:
            fig.savefig(
                self.output_dir
                / self.__class__.__name__
                / f"topomap_{self.__class__.__name__[:-4].lower()}_{prop.lower()}.png"
            )

    def plot_topomap_collage(
        self,
        props: Iterable[str],
        aggfunc: str = "mean",
        stages_to_plot: tuple = "all",
        sleep_stages: dict = {"Wake": 0, "N1": 1, "N2": 2, "N3": 3, "REM": 4},
        low_percentile: float = 5,
        high_percentile: float = 95,
        save: bool = False,
        topomap_args: dict = None,
        cbar_args: dict = None,
        figure_args: dict = None,
    ):
        """Plots topomap collage for multiple sleep stages and event properties.

        Args:
            props: Properties from the self.results.summary() to generate topomaps for.
            aggfunc: Averaging function, "mean" or "median". Defaults to "mean".
            stages_to_plot: stages_to_plot: Tuple of strings representing names from sleep_stages,
                e.g., ("REM", "N1"). If set to "all" plots every stage provided in sleep_stages.
                Defaults to "all".
            sleep_stages: Mapping between sleep stages names and their integer representations.
                Defaults to {"Wake": 0, "N1": 1, "N2": 2, "N3": 3, "REM": 4}.
            low_percentile: Set min color value by percentile of the property data.
                Defaults to 5.
            high_percentile: Set max color value by percentile of the property data.
                Defaults to 95.
            save: Whether to save the figure. Defaults to False.
            topomap_args: Arguments passed to `mne.viz.plot_topomap() <https://mne.tools/
                stable/generated/mne.viz.plot_topomap.html>`_. Defaults to None.
            cbar_args: Arguments passed to `plt.colorbar() <https://matplotlib.org/stable
                /api/_as_gen/matplotlib.pyplot.colorbar.html>`_. Defaults to None.
            figure_args: Arguments passed to `plt.figure() <https://matplotlib.org/stable/
                api/_as_gen/matplotlib.pyplot.figure.html>`_. Defaults to None.

        """
        from natsort import natsort_keygen
        from more_itertools import collapse

        topomap_args = topomap_args or dict()
        cbar_args = cbar_args or dict()
        figure_args = figure_args or dict()
        topomap_args.setdefault("cmap", "plasma")
        topomap_args.setdefault("vlim", [None, None])

        if stages_to_plot == "all":
            stages_to_plot = {
                k: v
                for k, v in sleep_stages.items()
                if v in self.results.summary()["Stage"].unique()
            }
        n_rows = len(stages_to_plot)
        n_cols = len(props)

        figure_args.setdefault("figsize", (n_cols * 4, n_rows * 4))
        figure_args.setdefault("layout", "constrained")
        fig = plt.figure(**figure_args)

        subfigs = fig.subfigures(n_rows, 1)

        grouped_summary = (
            self.results.summary(grp_chan=True, grp_stage=True, aggfunc=aggfunc)
            .sort_values("Channel", key=natsort_keygen())
            .reset_index()
        )

        if low_percentile:
            perc_high = dict()
        if high_percentile:
            perc_low = dict()

        data_per_stage_per_prop = np.empty(
            (len(stages_to_plot), len(props)), dtype=object
        )
        info = np.empty((len(stages_to_plot), len(props)), dtype=object)

        for col_index, prop in enumerate(props):
            for_perc = []
            for row_index, stage in enumerate(stages_to_plot):
                per_stage = grouped_summary.loc[
                    grouped_summary["Stage"].isin(collapse([sleep_stages[stage]]))
                ].groupby("Channel")
                per_stage = (
                    per_stage.mean() if aggfunc == "mean" else per_stage.median()
                )
                per_stage = per_stage.sort_values(
                    "Channel", key=natsort_keygen()
                ).reset_index()

                info[row_index, col_index] = (
                    self.mne_raw.copy().pick(list(per_stage["Channel"].unique())).info
                )
                data = per_stage[prop]
                for_perc.append(data.to_numpy())
                data_per_stage_per_prop[row_index, col_index] = data
            if low_percentile:
                perc_low[prop] = np.percentile(np.concatenate(for_perc), low_percentile)
            if high_percentile:
                perc_high[prop] = np.percentile(
                    np.concatenate(for_perc), high_percentile
                )

        for row_index, stage in enumerate(stages_to_plot):
            axes = subfigs[row_index].subplots(1, n_cols)

            for col_index, prop in enumerate(props):
                if low_percentile:
                    topomap_args["vlim"][0] = perc_low[prop]
                if high_percentile:
                    topomap_args["vlim"][1] = perc_high[prop]
                self._plot_topomap(
                    data=data_per_stage_per_prop[row_index, col_index],
                    axis=axes[col_index],
                    info=info[row_index, col_index],
                    topomap_args=topomap_args,
                    cbar_args=cbar_args,
                )
                axes[col_index].set_title(f"{prop}")
            n_spindles = int(
                self.results.summary(grp_chan=False, grp_stage=True).loc[
                    sleep_stages[stage]
                ]["Count"]
            )
            subfigs[row_index].suptitle(
                f"{stage}, {n_spindles} {self.__class__.__name__[:-4].lower()}",
                fontsize="xx-large",
            )

        fig.suptitle(f"{self.__class__.__name__[:-4]}", fontsize="xx-large")
        if save:
            fig.savefig(
                self.output_dir
                / self.__class__.__name__
                / f"topomap_{self.__class__.__name__[:-4].lower()}_collage.png"
            )

    def apply_tfr(
        self,
        freqs: Iterable[float],
        n_freqs: int,
        time_before: float,
        time_after: float,
        method: str = "morlet",
        save: bool = False,
        tfr_method_args: dict = None,
    ):
        """Transforms the events signal to time-frequency representation.

        Args:
            freqs: Lower and upper bounds of frequencies of interest in Hz, e.g., (10,20)
            n_freqs: Frequency resolution in TFR.
            time_before: Seconds before the event peak to get from the real data.
            time_after: Seconds after the event peak to get from the real data
            method: TFR transform method. Defaults to "morlet".
            save: Whether to save the TFRs to file. Defaults to False.
            tfr_method_args: Arguments passed to `mne.time_frequency.tfr_array_morlet()
                <https://mne.tools/stable/generated/mne.time_frequency.tfr_array_morlet.html>`_
                or `mne.time_frequency.tfr_array_multitaper() <https://mne.tools/stable/
                generated/mne.time_frequency.tfr_array_multitaper.html>`_. Defaults to None.
        """
        assert self.results, "Run detect method first"
        assert (
            method == "morlet" or method == "multitaper"
        ), "method should be 'morlet' or 'multitaper'"
        from natsort import natsorted

        sleep_stages = {
            -2: "Unscored",
            -1: "Art",
            0: "Wake",
            1: "N1",
            2: "N2",
            3: "N3",
            4: "REM",
        }

        tfr_method_args = tfr_method_args or dict()
        tfr_method_args.setdefault("n_jobs", -1)
        tfr_method_args.setdefault("verbose", "error")
        tfr_method_args["output"] = "avg_power"

        freqs = np.linspace(freqs[0], freqs[1], n_freqs)
        df_raw = self.results.get_sync_events(
            time_before=time_before, time_after=time_after
        )[["Event", "Amplitude", "Channel", "Stage"]]

        self.tfrs = {}

        for stage in df_raw["Stage"].unique():
            df = df_raw[df_raw["Stage"] == stage]
            # Group amplitudes by channel and events
            df = df.groupby(["Channel", "Event"], group_keys=True).apply(
                lambda x: x["Amplitude"]
            )

            # Create dict for every channel and values with array of shape (n_events, 1, n_event_times)
            for_tfrs = {
                channel: np.expand_dims(
                    np.array(
                        events_df.groupby(level=1)
                        .apply(lambda x: x.to_numpy())
                        .tolist()
                    ),
                    axis=1,
                )
                for channel, events_df in df.groupby(level=0)
            }

            # Calculate tfrs
            if method == "morlet":
                tfrs = {
                    channel: mne.time_frequency.tfr_array_morlet(
                        v,
                        self.sf,
                        freqs,
                        **tfr_method_args,
                    )
                    for channel, v in tqdm(for_tfrs.items())
                }
            elif method == "multitaper":
                tfrs = {
                    channel: mne.time_frequency.tfr_array_multitaper(
                        v,
                        self.sf,
                        freqs,
                        **tfr_method_args,
                    )
                    for channel, v in tqdm(for_tfrs.items())
                }
            # Sort and combine
            data = np.squeeze(
                np.array([tfr for channel, tfr in natsorted(tfrs.items())])
            )
            if data.ndim == 2:
                data = np.expand_dims(data, axis=0)
            self.tfrs[sleep_stages[stage]] = mne.time_frequency.AverageTFR(
                info=self.mne_raw.copy().pick(list(tfrs.keys())).info,
                data=data,
                times=np.linspace(
                    -time_before,
                    time_after,
                    int((time_before + time_after) * self.sf + 1),
                ),
                freqs=freqs,
                nave=np.mean([arr.shape[0] for arr in for_tfrs.values()], dtype=int),
                method=method,
            )
        if save:
            for stage, tfr in self.tfrs.items():
                tfr.save(
                    self.output_dir
                    / self.__class__.__name__
                    / f"{self.__class__.__name__[:-4].lower()}_{stage}-tfr.h5"
                )

    def read_tfrs(self, dirpath=None):
        """Loads TFRs stored in hdf5 files.

        Filenames should end with {type_of_event}_{sleep_stage}-tfr.h5

        Args:
            dirpath: Path to the directory containing hdf5 files. Defaults to None.
        """
        from mne.time_frequency import read_tfrs
        from pathlib import Path
        import re

        r = f"{self.__class__.__name__[:-4].lower()}_(.+)(?:-tfr.h5)"
        self.tfrs = dict()
        dirpath = (
            Path(dirpath) if dirpath else self.output_dir / self.__class__.__name__
        )
        for p in dirpath.glob("*tfr.h5"):
            m = re.search(r, str(p))
            if m:
                self.tfrs[m.groups()[0]] = read_tfrs(p)


@define(kw_only=True, slots=False)
class BaseSpectrum(BaseTopomap, ABC):
    """A template class for the spectral analysis."""

    psd_per_stage: dict = field(init=False)
    """ Dictionary of the form sleep_stage:[freqs array with shape (n_freqs), 
    psd array with shape (n_electrodes, n_freqs), 
    sleep_stage percent from the whole unrejected data]
    """

    def plot_psd_per_stage(
        self,
        picks: str | Iterable[str] = ("E101",),
        psd_range: tuple = (-40, 60),
        freq_range: tuple = (0, 40),
        xscale: str = "linear",
        sleep_stages: dict = {"Wake": 0, "N1": 1, "N2": 2, "N3": 3, "REM": 4},
        axis: plt.axis = None,
        plot_sensors: bool = False,
        save: bool = False,
        psd_method_args: dict = None,
        subplots_args: dict = None,
    ):
        """Plot PSD per sleep stage.

        Args:
            picks: Channels to calculate PSD on, more info at
                `mne.io.Raw.get_data <https://mne.tools/stable/generated/
                mne.io.Raw.html#mne.io.Raw.get_data>`_.
                Defaults to ("E101",).
            psd_range: Range of y axis on PSD plot. Defaults to (-40, 60).
            freq_range: Range of x axis on PSD plot. Defaults to (0, 40).
            xscale: Scale of the X axis, check available values at
                `matplotlib.axes.Axes.set_xscale <https://mne.tools/stable/
                generated/mne.io.Raw.html#mne.io.Raw.get_data>`_.
                Defaults to "linear".
            sleep_stages: Mapping between sleep stages names and their integer representations.
                Defaults to {"Wake": 0, "N1": 1, "N2": 2, "N3": 3, "REM": 4}.
            axis: Instance of `matplotlib.pyplot.axis <https://matplotlib.org/
                stable/api/_as_gen/matplotlib.pyplot.axis.html#matplotlib-pyplot-axis>`_.
                Defaults to None.
            plot_sensors: Whether to plot sensor map showing which channels were used for
                computing PSD. Defaults to False.
            save: Whether to save the figure. Defaults to False.
            psd_method_args: Arguments passed to the PSD method, e.g., welch. Defaults to None.
            subplots_args: Arguments passed to the plt.subplots(). Have no effect if axis is provided.
                Defaults to None.
        """
        subplots_args = subplots_args or dict()
        is_new_axis = False

        if not axis:
            fig, axis = plt.subplots(**subplots_args)
            is_new_axis = True

        psd_per_stage = self._compute_psd_per_stage(
            picks=picks,
            sleep_stages=sleep_stages,
            avg_ref=False,
            dB=True,
            method_args=psd_method_args,
        )

        for stage in sleep_stages:
            axis.plot(
                psd_per_stage[stage][0],
                psd_per_stage[stage][1][0],
                label=f"{stage} ({psd_per_stage[stage][2]}%)",
            )

        axis.set_xlim(freq_range)
        axis.set_ylim(psd_range)
        axis.set_xscale(xscale)
        axis.set_title("Welch's PSD")
        axis.set_ylabel("PSD [dB/Hz]")
        axis.set_xlabel(f"{xscale} frequency [Hz]".capitalize())
        axis.legend()

        if plot_sensors:
            from mne.io.pick import _picks_to_idx
            import matplotlib.patches as mpatches
            from mpl_toolkits.axes_grid1.inset_locator import inset_axes

            color = "cyan"
            # ax = fig.add_subplot(3, 3, 7)
            axins = inset_axes(axis, width="30%", height="30%", loc="lower left")
            channels = np.array(self.mne_raw.info.ch_names)[
                _picks_to_idx(self.mne_raw.info, picks)
            ].tolist()
            self._plot_sensors(
                ch_colors={ch: color for ch in channels},
                axes=axins,
            )

            patches = []
            patches.append(mpatches.Patch(color=color, label="psd"))
            if self.mne_raw.info["bads"]:
                patches.append(mpatches.Patch(color="red", label="bad"))
            axins.legend(
                handles=patches,
                loc="lower left",
                bbox_to_anchor=(1, 0),
                fontsize="x-small",
            )

        # Save the figure if 'save' set to True and no axis has been passed.
        if save and is_new_axis:
            fig.savefig(self.output_dir / self.__class__.__name__ / f"psd.png")

    def plot_topomap(
        self,
        stage: str = "REM",
        band: dict = {"Delta": (0, 4)},
        dB: bool = False,
        sleep_stages: dict = {"Wake": 0, "N1": 1, "N2": 2, "N3": 3, "REM": 4},
        axis: plt.axis = None,
        fooof: bool = False,
        save: bool = False,
        psd_method_args: dict = None,
        topomap_args: dict = None,
        cbar_args: dict = None,
        fooof_group_args: dict = None,
        fooof_get_band_peak_fg_args: dict = None,
        subplots_args: dict = None,
    ):
        """Plots topomap for a sleep stage and a frequency band.

        Args:
            stage: One of the sleep_stages keys. Defaults to "REM".
            band: Name-value pair - with name=arbitrary name
                and value=(l_freq, h_freq).
                Defaults to {"Delta": (0, 4)}.
            dB: Whether transform PSD to dB. Defaults to False.
            sleep_stages: Mapping between sleep stages names and their integer representations.
                Defaults to {"Wake": 0, "N1": 1, "N2": 2, "N3": 3, "REM": 4}.
            axis: Instance of `matplotlib.pyplot.axis <https://matplotlib.org/
                stable/api/_as_gen/matplotlib.pyplot.axis.html#matplotlib-pyplot-axis>`_.
                Defaults to None.
            fooof: Whether to plot parametrised spectra.
                More at `fooof <https://fooof-tools.github.io/fooof/auto_examples/analyses/
                plot_mne_example.html#sphx-glr-auto-examples-analyses-plot-mne-example-py>`_.
                Defaults to False.
            save: Whether to save the figure. Defaults to False.
            psd_method_args: Arguments passed to the PSD method, e.g., welch. Defaults to None.
            topomap_args: Arguments passed to `mne.viz.plot_topomap() <https://mne.tools/
                stable/generated/mne.viz.plot_topomap.html>`_. Defaults to None.
            cbar_args: Arguments passed to `plt.colorbar() <https://matplotlib.org/stable
                /api/_as_gen/matplotlib.pyplot.colorbar.html>`_. Defaults to None.
            fooof_group_args: Arguments passed to `fooof.FOOOFGroup() <https://fooof-tools.github.io/
                fooof/generated/fooof.FOOOFGroup.html#fooof.FOOOFGroup>`_. Defaults to None.
            fooof_get_band_peak_fg_args: Arguments passed to the `fooof.analysis.get_band_peak_fg()
                <https://fooof-tools.github.io/fooof/generated/fooof.analysis.get_band_peak_fg.html>`_.
                Defaults to None.
            subplots_args: Arguments passed to the plt.subplots(). Have no effect if axis is provided.
                Defaults to None.
        """
        topomap_args = topomap_args or dict()
        cbar_args = cbar_args or dict()
        subplots_args = subplots_args or dict()
        topomap_args.setdefault("cmap", "plasma")
        cbar_args.setdefault(
            "label", "dB/Hz" if dB else r"$\mu V^{2}/Hz$" if not fooof else None
        )
        assert (
            stage in sleep_stages.keys()
        ), f"sleep_stages should contain provided stage"

        is_new_axis = False

        if axis is None:
            fig, axis = plt.subplots(**subplots_args)
            is_new_axis = True

        self.psd_per_stage = self._compute_psd_per_stage(
            picks=["eeg"],
            sleep_stages=sleep_stages,
            avg_ref=True,
            dB=dB,
            method_args=psd_method_args,
        )

        [(_, b)] = band.items()

        if fooof:
            psds = self._fooof(
                band=band,
                stage=stage,
                fooof_group_args=fooof_group_args,
                get_band_peak_fg_args=fooof_get_band_peak_fg_args,
            )

        else:
            psds = np.take(
                self.psd_per_stage[stage][1],
                np.where(
                    np.logical_and(
                        self.psd_per_stage[stage][0] >= b[0],
                        self.psd_per_stage[stage][0] <= b[1],
                    )
                )[0],
                axis=1,
            ).sum(axis=1)

        self._plot_topomap(
            data=psds,
            axis=axis,
            topomap_args=topomap_args,
            cbar_args=cbar_args,
        )
        if is_new_axis:
            fig.suptitle(f"{stage} ({b[0]}-{b[1]} Hz)")
        if save and is_new_axis:
            fig.savefig(
                self.output_dir
                / self.__class__.__name__
                / f"topomap_psd_{list(band)[0]}.png"
            )

    def plot_topomap_collage(
        self,
        stages_to_plot: tuple = "all",
        bands: dict = {
            "Delta": (0, 3.99),
            "Theta": (4, 7.99),
            "Alpha": (8, 12.49),
            "SMR": (12.5, 15),
            "Beta": (12.5, 29.99),
            "Gamma": (30, 60),
        },
        dB: bool = False,
        sleep_stages: dict = {"Wake": 0, "N1": 1, "N2": 2, "N3": 3, "REM": 4},
        fooof: bool = False,
        low_percentile: float = 5,
        high_percentile: float = 95,
        save: bool = False,
        psd_method_args: dict = None,
        topomap_args: dict = None,
        cbar_args: dict = None,
        fooof_group_args: dict = None,
        fooof_get_band_peak_fg_args: dict = None,
        figure_args: dict = None,
    ):
        """Plots topomap collage for multiple sleep stages and bands.

        Args:
            stages_to_plot: Tuple of strings representing names from sleep_stages,
                e.g., ("REM", "N1").
                If set to "all" plots every stage provided in sleep_stages.
                Defaults to "all".
            bands: Dict of name-value pairs - with name=arbitrary name
                and value=(l_freq, h_freq).
                Defaults to { "Delta": (0, 3.99), "Theta": (4, 7.99), "Alpha": (8, 12.49),
                "SMR": (12.5, 15), "Beta": (12.5, 29.99), "Gamma": (30, 60), }.
            dB: Whether transform PSD to dB. Defaults to False.
            sleep_stages: Mapping between sleep stages names and their integer representations.
                Defaults to {"Wake": 0, "N1": 1, "N2": 2, "N3": 3, "REM": 4}.
            fooof: Whether to plot parametrised spectra.
                More at `fooof <https://fooof-tools.github.io/fooof/auto_examples/analyses/
                plot_mne_example.html#sphx-glr-auto-examples-analyses-plot-mne-example-py>`_.
                Defaults to False.
            low_percentile: Set min color value by percentile of the band data.
                Defaults to 5.
            high_percentile: Set max color value by percentile of the band data.
                Defaults to 95.
            save: Whether to save the figure. Defaults to False.
            psd_method_args: Arguments passed to the PSD method, e.g., welch. Defaults to None.
            topomap_args: Arguments passed to `mne.viz.plot_topomap() <https://mne.tools/
                stable/generated/mne.viz.plot_topomap.html>`_. Defaults to None.
            cbar_args: Arguments passed to `plt.colorbar() <https://matplotlib.org/stable
                /api/_as_gen/matplotlib.pyplot.colorbar.html>`_. Defaults to None.
            fooof_group_args: Arguments passed to `fooof.FOOOFGroup() <https://fooof-tools.github.io/
                fooof/generated/fooof.FOOOFGroup.html#fooof.FOOOFGroup>`_. Defaults to None.
            fooof_get_band_peak_fg_args: Arguments passed to the `fooof.analysis.get_band_peak_fg()
                <https://fooof-tools.github.io/fooof/generated/fooof.analysis.get_band_peak_fg.html>`_.
                Defaults to None.
            figure_args: Arguments passed to `plt.figure() <https://matplotlib.org/stable/
                api/_as_gen/matplotlib.pyplot.figure.html>`_. Defaults to None.
        """
        topomap_args = topomap_args or dict()
        cbar_args = cbar_args or dict()
        figure_args = figure_args or dict()
        topomap_args.setdefault("cmap", "plasma")
        topomap_args.setdefault("vlim", [None, None])
        cbar_args.setdefault(
            "label", "dB/Hz" if dB else r"$\mu V^{2}/Hz$" if not fooof else None
        )

        if stages_to_plot == "all":
            stages_to_plot = sleep_stages.keys()
        n_rows = len(stages_to_plot)
        n_cols = len(bands)

        figure_args.setdefault("figsize", (n_cols * 4, n_rows * 4))
        figure_args.setdefault("layout", "constrained")
        fig = plt.figure(**figure_args)
        subfigs = fig.subfigures(n_rows, 1)

        self.psd_per_stage = self._compute_psd_per_stage(
            picks=["eeg"],
            sleep_stages=sleep_stages,
            avg_ref=True,
            dB=dB,
            method_args=psd_method_args,
        )

        if low_percentile:
            perc_low = dict()

        if high_percentile:
            perc_high = dict()

        psds_per_stage_per_band = np.empty(
            (len(stages_to_plot), len(bands)), dtype=object
        )
        for col_index, (band_key, b) in enumerate(bands.items()):
            for_perc = []
            for row_index, stage in enumerate(self.psd_per_stage):
                if fooof:
                    psds = self._fooof(
                        band={band_key: b},
                        stage=stage,
                        fooof_group_args=fooof_group_args,
                        get_band_peak_fg_args=fooof_get_band_peak_fg_args,
                    )
                else:
                    psds = np.take(
                        self.psd_per_stage[stage][1],
                        np.where(
                            np.logical_and(
                                self.psd_per_stage[stage][0] >= b[0],
                                self.psd_per_stage[stage][0] <= b[1],
                            )
                        )[0],
                        axis=1,
                    ).sum(axis=1)
                for_perc.append(psds)
                psds_per_stage_per_band[row_index, col_index] = psds
            if low_percentile:
                perc_low[band_key] = np.percentile(for_perc, low_percentile)
            if high_percentile:
                perc_high[band_key] = np.percentile(for_perc, high_percentile)

        for row_index, stage in enumerate(stages_to_plot):
            axes = subfigs[row_index].subplots(1, n_cols)

            for col_index, band_key in enumerate(bands):
                if low_percentile:
                    topomap_args["vlim"][0] = perc_low[band_key]
                if high_percentile:
                    topomap_args["vlim"][1] = perc_high[band_key]

                self._plot_topomap(
                    data=psds_per_stage_per_band[row_index, col_index],
                    axis=axes[col_index],
                    topomap_args=topomap_args,
                    cbar_args=cbar_args,
                )
                axes[col_index].set_title(
                    f"{band_key} ({bands[band_key][0]}-{bands[band_key][1]} Hz)"
                )

            subfigs[row_index].suptitle(
                f"{stage} ({self.psd_per_stage[stage][2]}%)", fontsize="xx-large"
            )

        if save:
            fig.savefig(
                self.output_dir / self.__class__.__name__ / f"topomap_psd_collage.png"
            )

    def _fooof(self, band, stage, fooof_group_args=None, get_band_peak_fg_args=None):
        from fooof import FOOOFGroup
        from fooof.bands import Bands
        from fooof.analysis import get_band_peak_fg

        fooof_group_args = fooof_group_args or dict()
        get_band_peak_fg_args = get_band_peak_fg_args or dict()
        fooof_group_args.setdefault("peak_width_limits", (1, 6))
        fooof_group_args.setdefault("min_peak_height", 0.15)
        fooof_group_args.setdefault("peak_threshold", 2.0)
        fooof_group_args.setdefault("verbose", False)
        # Initialize a FOOOFGroup object, with desired settings
        fg = FOOOFGroup(**fooof_group_args)

        # Define the frequency range to fit
        freq_range = [1, 45]

        fg.fit(
            self.psd_per_stage[stage][0],
            self.psd_per_stage[stage][1],
            freq_range=freq_range,
        )

        # Define frequency bands of interest
        bands = Bands(band)

        # Extract peaks
        peaks = get_band_peak_fg(fg, bands[list(band)[0]], **get_band_peak_fg_args)

        peaks[np.where(np.isnan(peaks))] = 0
        # Extract the power values from the detected peaks
        psds = peaks[:, 1]

        return psds

    @abstractmethod
    def _compute_psd_per_stage(self):
        pass



from setuptools import setup

setup(
    name="sleepeeg",
    version="0.0.1",
    install_requires=[
        "attrs",
        "mne",
        "mne-qt-browser",
        "pyqt5",
        "yasa==0.6.3",
        "lspopt",
    ],
)
